@@ -5,12 +5,10 @@ import numpy as np
 from numpy.typing import ArrayLike
 import requests
 import time
-import trimesh
-from typing import Callable, Collection, Iterable, Optional, Union
+from typing import Callable, Collection, Iterable, Optional, Tuple, Union
 from werkzeug.datastructures import ImmutableMultiDict
 
 from .auth import get_credentials
-from .bracelet_electrodes import generate_bracelet_electrodes as generate_bracelet_electrode_centers
 
 logger = logging.getLogger(__name__)
 
@@ -39,6 +37,16 @@ class SurfaceType(IntEnum):
     ELECTRODE = 4
 
 
+class ElectrodeLayouts(IntEnum):
+    """Possible orderings for grids of electrodes on the skin"""
+    ROW_MAJOR = 0
+    COLUMN_MAJOR = auto()
+    SNAKE_HORIZONTAL = auto()
+    SNAKE_VERTICAL = auto()
+    SPIRAL_CLOCKWISE = auto()
+    SPIRAL_COUNTERCLOCKWISE = auto()
+
+
 # Random seed to be used in new() calls when not explicitly given as parameter.
 DEFAULT_RANDOM_SEED = 0
 
@@ -486,55 +494,146 @@ class Electrode(MDTObject):
         return self._location.copy()
 
 
-def generate_bracelet_electrodes(skin: Surface,
-                                 radius: float,
-                                 first_electrode_location: ArrayLike,
-                                 rings_normal: ArrayLike,
-                                 n_rings: int,
-                                 distance_between_rings: float,
-                                 n_electrodes_per_ring: int,
-                                 distance_between_electrodes: Optional[float] = None):
-    """Generate a bracelet of electrodes on the given surface.
+class ElectrodeCollection(MDTObject):
+    _electrodes: Optional[Collection[Electrode]] = None
+    _radius: Optional[float] = None
 
-    A bracelet is a series of electrodes arranged in multiple rings that should wrap around the given surface.
+    @classmethod
+    def _send_request_and_set_radius(cls, json, radius):
+        # Send the request, then set the radius and then return the metadata instance.
+        metadata = cls._new_request(json=json)
+        metadata._radius = radius
+        return metadata
 
-    Args:
-        skin: A Surface object with type OUTER_SKIN where the electrodes will be placed.
-        radius: Radius of an electrode. The same value is used for all electrodes in the bracelet.
-        first_electrode_location: A numpy array containing 3D coordinates of the first electrode of the first ring.
-            Does not have to be precisely on the skin mesh - the point will be projected onto it.
-        rings_normal: A numpy array containing the direction of the normal vector defining the bracelet section planes.
-            Note, that all the rings have the same normal vector.
-        n_rings: The number of electrode rings in the bracelet. If larger than 1, the initial bracelet plan is shifted
-            in the direction of the normal vector for a distance distance_between_rings.
-        distance_between_rings: Distance between electrode rings in the bracelet, in meters. Note, that all the rings
-            are equidistant.
-        n_electrodes_per_ring: The number of electrodes in rings. The electrodes are ordered in a clockwise direction
-            with respect to the rings_normal. If the distance_between_rings parameter is 0, all the rings will have the
-            same number of electrodes uniformly covering the corresponding section circumferences. It will likely
-            result in different inter electrode distances for different rings.
-        distance_between_electrodes: The distance between electrodes in a ring, in meters. If not None, will generate
-            equidistant in a clockwise direction starting from the first_electrode_location.
-    """
-    # Make sure the surface is indeed the skin.
-    if skin.type != SurfaceType.OUTER_SKIN:
-        raise ValueError(f"The provided Surface object shouild be of type 'OUTER_SKIN', not '{skin.type.name}'.")
-    centers = generate_bracelet_electrode_centers(trimesh.Trimesh(vertices=skin.vertices, faces=skin.triangles),
-                                                  first_electrode_location,
-                                                  rings_normal,
-                                                  n_rings,
-                                                  distance_between_rings,
-                                                  n_electrodes_per_ring,
-                                                  distance_between_electrodes
-                                                  ).reshape(-1, 3)
-    # For each electrode center, create a new electrode and save it into a list that is returned to the user.
-    electrodes = []
-    for c in centers:
-        new_electrode = Electrode.new(c, radius)
-        if wait:
-            new_electrode.wait()
-        electrodes.append(new_electrode)
-    return electrodes
+    @property
+    def electrodes(self):
+        if self._electrodes is None:
+            self._retrieve_data()
+            electrode_locations = _decode_array(self._data_json, "electrode_locations")
+            self._electrodes = [Electrode.new(location, self._radius) for location in electrode_locations]
+
+        return self._electrodes
+
+
+class ElectrodeBracelet(ElectrodeCollection):
+    _url_prefix = "electrode_bracelets"
+
+    @classmethod
+    def new(
+            cls,
+            skin: Surface,
+            radius: float,
+            first_electrode_location: ArrayLike,
+            rings_normal: ArrayLike,
+            n_rings: int,
+            distance_between_rings: float,
+            n_electrodes_per_ring: int,
+            distance_between_electrodes: Optional[float] = None,
+            force_computation: bool = False
+    ) -> ElectrodeBracelet:
+        """Generate a bracelet of electrodes on the given surface.
+
+        A bracelet is a series of electrodes arranged in multiple rings that should wrap around the given surface.
+
+        Args:
+            skin: A Surface object with type OUTER_SKIN where the electrodes will be placed.
+            radius: Radius of an electrode. The same value is used for all electrodes in the bracelet.
+            first_electrode_location: A numpy array containing 3D coordinates of the first electrode of the first ring.
+                Does not have to be precisely on the skin mesh - the point will be projected onto it.
+            rings_normal: A numpy array containing the direction of the normal vector defining the bracelet section planes.
+                Note, that all the rings have the same normal vector.
+            n_rings: The number of electrode rings in the bracelet. If larger than 1, the initial bracelet plan is shifted
+                in the direction of the normal vector for a distance distance_between_rings.
+            distance_between_rings: Distance between electrode rings in the bracelet, in meters. Note, that all the rings
+                are equidistant.
+            n_electrodes_per_ring: The number of electrodes in rings. The electrodes are ordered in a clockwise direction
+                with respect to the rings_normal. If the distance_between_rings parameter is 0, all the rings will have the
+                same number of electrodes uniformly covering the corresponding section circumferences. It will likely
+                result in different inter electrode distances for different rings.
+            distance_between_electrodes: The distance between electrodes in a ring, in meters. If not None, will generate
+                equidistant in a clockwise direction starting from the first_electrode_location.
+            force_computation: Boolean flag that allows to bypass a status check. Normally, the API would check if an
+                object with the requested parameters was already present in the database. If this was the case, the API
+                would not request its computation and would simply return its identifier. By setting this flag to
+                `True`, the user can bypass this logic and force the computation of the resource even if it existed
+                already. The main reason is to allow starting a calculation again in case of unexpected errors.
+        """
+        # Encode the parameters.
+        json = {
+            "surface_id": skin.id,
+            "n_rings": n_rings,
+            "distance_between_rings": distance_between_rings,
+            "n_electrodes_per_ring": n_electrodes_per_ring,
+            "distance_between_electrodes": distance_between_electrodes,
+            "force_computation": force_computation,
+        }
+        json.update(_encode_array(first_electrode_location, "first_electrode_location"))
+        json.update(_encode_array(rings_normal, "rings_normal"))
+        return cls._send_request_and_set_radius(json, radius)
+
+
+class ElectrodeGrid(ElectrodeCollection):
+    _url_prefix = "electrode_grids"
+
+    @classmethod
+    def new(
+            cls,
+            skin: Surface,
+            radius: float,
+            corners: ArrayLike,
+            shape: Tuple[int, int],
+            layout: ElectrodeLayouts,
+            smooth_surface: bool = True,
+            force_computation: bool = False
+    ) -> ElectrodeGrid:
+        """Generate a grid of electrodes on the given surface.
+
+        Args:
+            skin: A Surface object with type OUTER_SKIN where the electrodes will be placed.
+            radius: Radius of an electrode. The same value is used for all electrodes in the bracelet.
+            corners: A NumPy array with shape (4, 3) representing the 3D corners of a rectangular patch of skin that
+                should contain a grid of electrodes. The four corners are to be given in the order top-left (grid[0]),
+                top-right (grid[1]), bottom-left (grid[2]), bottom-right (grid[3]). The points do not need to define an
+                actual rectangle, but rather an approximation of it. In addition, they do not have to be precisely on
+                the skin mesh.
+            shape: A NumPy array with shape (2,), containing the rows and columns (respectively as first and second
+                element) of the grid of electrodes.
+            layout: Ordering of the electrodes. The final result - accessible via the `electrodes` property - is a list
+                of electrodes and it is therefore necessary to specify how to order them. The allowed values have the
+                following meanings:
+                - ROW_MAJOR: electrodes are rearranged row-by-row.
+                - COLUMN_MAJOR: electrodes are rearranged column-by-column.
+                - SNAKE_HORIZONTAL: like ROW_MAJOR, but every odd-numbered row is reversed.
+                - SNAKE_VERTICAL: like COLUMN_MAJOR, but every odd-numbered column is reversed.
+                - SPIRAL_CLOCKWISE: creates a spiral path starting with the first row of the grid and proceeding in
+                    clockwise order.
+                - SPIRAL_COUNTERCLOCKWISE: creates a spiral path starting with the first column of the grid and proceeding
+                    in counterclockwise order.
+            smooth_surface: If True (default), use a smoothed version of the skin surface during the evaluation of the
+                grid. This generally provides more regular grids. The final grid is still projected on the original
+                surface mesh.
+            force_computation: Boolean flag that allows to bypass a status check. Normally, the API would check if an
+                object with the requested parameters was already present in the database. If this was the case, the API
+                would not request its computation and would simply return its identifier. By setting this flag to
+                `True`, the user can bypass this logic and force the computation of the resource even if it existed
+                already. The main reason is to allow starting a calculation again in case of unexpected errors.
+        """
+        # Check that parameters are correctly set.
+        corners = np.asarray(corners)
+        if corners.shape != (4, 3):
+            raise ValueError(f"Argument 'corners' has shape {corners.shape}. Expected: (4, 3).")
+
+        # Encode the parameters.
+        json = {
+            "surface_id": skin.id,
+            "rows": shape[0],
+            "cols": shape[1],
+            "layout": layout.value,
+            "smooth_surface": smooth_surface,
+            "force_computation": force_computation,
+        }
+        json.update(_encode_array(corners.reshape(-1), "grid_corners"))
+        return cls._send_request_and_set_radius(json, radius)
 
 
 class ForwardSolution(MDTObject):
@@ -639,13 +738,13 @@ class Fibers(MDTObject):
             building new ones.
         """
         # Convert the planes to numpy arrays and verify the shape.
-        plane_origins = np.array(plane_origins, dtype=np.float)
+        plane_origins = np.array(plane_origins, dtype=np.float32)
         if plane_origins.shape != (2, 3):
             raise ValueError(
                 "The plane origins must be a 2D array of floats with a shape of (2, 3),"
                 f" not {plane_origins.shape}."
             )
-        plane_normals = np.array(plane_normals, dtype=np.float)
+        plane_normals = np.array(plane_normals, dtype=np.float32)
         if plane_normals.shape != (2, 3):
             raise ValueError(
                 "The plane normals must be a 2D array of floats with a shape of (2, 3),"
@@ -1063,7 +1162,7 @@ class Electromyography(MDTObject):
                                     json={"credentials": get_credentials()})
             if not response.ok:
                 raise ValueError("The EMG data could not be retrieved.")
-            self._data = np.array(response.json()["data"], dtype=np.float)
+            self._data = np.array(response.json()["data"], dtype=np.float32)
 
         return self._data
 
